classdef Alignment
    properties (SetAccess = private)
        seqs        
        cons_cols
    end
    properties (GetAccess = private, SetAccess = private)
        sum_num_residuals
    end
    
    properties (Dependent = true)
        seq_len
        num_seqs
        subs
    end
    
    methods
        function al = Alignment(input_file_name, scenario)
            global all_subs;
            OUTPUT = true;
            % read sequence data from fasta file
            fid = fopen(input_file_name);
            if fid == -1 
                error(['Unable to open: ' input_file_name]) 
            end
            [st, rs] = system(['grep -c ">" ' input_file_name]);
            num_seqs = str2num(rs(1:end-1));
            seq_names = cell(1, num_seqs);
            genomes = cell(1, num_seqs);
            subs = cell(1, num_seqs);
            curr_genome = '';
            is_first = true;
            counter = 0;
            while 1
                tline = fgetl(fid);
                if tline == -1, break, end
                if (tline(1) == '>')
                    if is_first
                        is_first = false;
                    else
                        genomes{counter} = curr_genome;
                        curr_genome = '';
                        if OUTPUT
                            disp(['read seq ' num2str(counter)])
                        end
                    end
                    counter = counter + 1;
                    curr_seq_name = tline(2:end);
                    seq_names{counter} = curr_seq_name;
                    s = regexp(curr_seq_name, '(\.|_)');
                    if (isempty(s))
                        error(['Sequence name has wrong format: ' curr_seq_name])
                    end
                    if (isstrprop(curr_seq_name(1), 'alpha'))
                        subs{counter} = curr_seq_name(1); 
                    else
                        subs{counter} = curr_seq_name(1:s(1)-1); 
                    end
                else
                    if is_first
                        error('First character in alignment file is not ">"')
                    end
                    curr_genome = [curr_genome, tline];
                end
            end
            genomes{counter} = curr_genome;
            fclose(fid);
            disp(['read ' num2str(num_seqs) ' sequences from ' input_file_name]);
            
            % generate seqs
            curr_seqs = [];
            seq_lens = [];
            for i=1:num_seqs
                curr_seqs = [curr_seqs Seq(seq_names{i}, genomes{i}, subs{i})];
                seq_lens(i) = curr_seqs(i).len();
                if OUTPUT
                    disp(['generated seq ' num2str(i)])
                end
           end

            % remove seqs with wrong sub
            valid_idx = [];
            for i=1:num_seqs
                if (all_subs.is_valid(curr_seqs(i).sub))
                    valid_idx = [valid_idx, i];
                end
                if OUTPUT
                    disp(['checked correct sub of seq ' num2str(i)])
                end
            end
            curr_seqs = curr_seqs(valid_idx);
            num_seqs = length(curr_seqs);
                       
            % find most frequent seq len
            seq_lens_unique = unique(seq_lens);
            check=@(c)sum(c == seq_lens);
            counts = arrayfun(check, seq_lens_unique); 
            [C,I] = max(counts);
            most_freq_len = seq_lens_unique(I);
            if OUTPUT
               disp('found most frequent seq length')
            end
           
            % remove seqs with wrong len
            valid_idx = [];
            for i=1:num_seqs
                if (curr_seqs(i).len ~= most_freq_len)
                    disp(['Sequence ' curr_seqs(i).name ' has been removed (correct length: ' num2str(most_freq_len) ', actual length: ' num2str(curr_seqs(i).len) ')'])
                else
                    valid_idx = [valid_idx, i];
                end
                if OUTPUT
                    disp(['checked length of seq ' num2str(i)])
                end
            end
            curr_seqs = curr_seqs(valid_idx);
            num_seqs = length(curr_seqs);
            
            % add seqs to all_subs
            for i=1:num_seqs
                all_subs.add_seq(curr_seqs(i));
                if OUTPUT
                    disp(['added seq ' num2str(i) ' to all_subs'])
                end
            end
            
            % display number of seqs
            for i=1:all_subs.num
                name = char(all_subs.names(i));
                num = all_subs.num_seqs(char(all_subs.names(i)));
                disp(['Number of valid sequences read of subtype ' name ': ' num2str(num)]);
            end

            al.seqs = curr_seqs;
            
            % calculate consensus cols
            M = [];
            for i=1:al.num_seqs()
                M = strvcat(M, al.seqs(i).genome);
            end
            al.cons_cols = ~all(Alpha.is_gap(M));
            if OUTPUT
                disp('calculated consensus columns')
            end

            % calculate number of residuals
            res_per_pos = sum(~Alpha.is_gap(M));
            s = 0;
            for i=1:al.seq_len
                s = s + res_per_pos(i);
                al.sum_num_residuals(i) = s;
                if OUTPUT
                    disp(['calculated number of residuals for pos ' num2str(i)])
                end
            end

            al.print_calc_etp_al(scenario);
        end
        
        function seq_len = get.seq_len(obj)
            seq_len = length(obj.seqs(1).genome);
        end        
        function num_seqs = get.num_seqs(obj)
            num_seqs = length(obj.seqs);
        end
        
        % present subs
        function subs = get.subs(obj)
            subs = {};
            for i=1:obj.num_seqs
                if (isempty(find(strcmp(subs, obj.seqs(i).sub))))
                    s = size(subs);
                    subs{s(2) + 1} = obj.seqs(i).sub;
                end                
            end
        end
        
        % numbers of all sequences
        function [no_seqs] = no_seqs(obj)
            no_seqs = 1:obj.num_seqs;
        end
                       
        % save to DB
        function save_to_db(obj, al_id)
            DB_con = SQL.get_connection();
            sql_stat = ['DELETE FROM Alignment WHERE ID=' num2str(al_id)];                 
            DB_con.executeSQLStatement(sql_stat);                    

            subs = obj.subs
            s = size(subs);
            for i=1:s(2)
                curr_no_seqs = obj.no_seqs_of_sub(subs(i));
                for j=1:size(curr_no_seqs)   
                    sql_query = ['select * from Seqs where Name=''' obj.seqs(curr_no_seqs(j)).name ''''];
                    rs = DB_con.executeSQLQuery(sql_query);                    
                    if (~rs.next())
                        sql_stat = ['INSERT INTO Seqs VALUES (''' obj.seqs(curr_no_seqs(j)).name ''', ''' subs{i} ''', ' ...
                                    num2str(j) ', ''' obj.seqs(curr_no_seqs(j)).genome ''')'];                 
                        DB_con.executeSQLStatement(sql_stat);                    
                    end                    
                    sql_stat = ['INSERT INTO Alignment VALUES (' num2str(al_id) ', ''' obj.seqs(curr_no_seqs(j)).name ''')'];
                    DB_con.executeSQLStatement(sql_stat);                                        
                end
            end
        end
        
        
        % get number of residuals in a given segment
        function num_res = num_residuals(obj, begin_pos, end_pos)
            if (begin_pos > 1)
                num_res = obj.sum_num_residuals(end_pos) - obj.sum_num_residuals(begin_pos - 1);
            else
                num_res = obj.sum_num_residuals(end_pos);
            end
        end        
        
        % numbers of the sequences belonging to a given sub
        function [no_seqs] = no_seqs_of_sub(obj, sub)
            seqs = mat2cell(obj.seqs, 1, ones(1, length(obj.seqs)));
            csub = @(a) a.sub;
            subs = char(cellfun(csub, seqs, 'UniformOutput', false));
            no_seqs = strmatch(sub, subs);
        end
        
        % calculate emission freqs
        function [freqs] = em_freqs(obj, begin_sector, end_sector, no_used_seqs) 
            global consts;    
            sector_len = end_sector - begin_sector + 1;
            freqs = zeros(sector_len, consts.alphabet.num_nucl);
            for i=begin_sector:end_sector
                for j=1:length(no_used_seqs)
                    char = obj.seqs(no_used_seqs(j)).genome(i);
                    if (~Alpha.is_gap(char))
                        idx = consts.alphabet.char2idx(char);
                        freqs(i - begin_sector + 1, idx) = freqs(i - begin_sector + 1, idx) + 1;
                    end
                end
            end
        end
        
        % calculate transition freqs
        function [freqs] = tr_freqs(obj, begin_sector, end_sector, no_used_seqs) 
            num_states = 2;
            sector_len = end_sector - begin_sector + 1;
            
            freqs = zeros(sector_len, num_states, num_states);
            i = begin_sector;
            while ((~obj.cons_cols(i)) && (i <= end_sector))
                i = i + 1;
            end
            
            while (i < end_sector)
                j = i + 1;
                while ((~obj.cons_cols(j)) && (j <= end_sector))
                    j = j + 1;
                end
                if (j <= end_sector)
                    for k=1:length(no_used_seqs)
                        source_is_delete = Alpha.is_gap(obj.seqs(k).genome(i)); 
                        dest_is_delete = Alpha.is_gap(obj.seqs(k).genome(j)); 

                        if (source_is_delete)
                            if (dest_is_delete)
                                freqs(i - begin_sector + 1, 2, 2) = freqs(i - begin_sector + 1, 2, 2) + 1;
                            else
                                freqs(i - begin_sector + 1, 2, 1) = freqs(i - begin_sector + 1, 2, 1) + 1;
                            end
                        else
                            if (dest_is_delete)
                                freqs(i - begin_sector + 1, 1, 2) = freqs(i - begin_sector + 1, 1, 2) + 1;
                            else
                                freqs(i - begin_sector + 1, 1, 1) = freqs(i - begin_sector + 1, 1, 1) + 1;
                            end
                        end
                    end
                end
                i = j;
            end
        end       
        
        function print_jpHMM_al(obj, begin_sector, end_sector, no_used_seqs, stamp)
           global all_subs;
           fid = fopen(FilesDirs.jpHMM_al_filename(stamp), 'w');
           if fid == -1 
               error(['Unable to open: ' FilesDirs.jpHMM_al_filename(stamp)]) 
           end
           for i=1:all_subs.num
                if (all_subs.num_seqs(all_subs.names{i}) && (length(all_subs.names{i}) == 1))
                    fprintf(fid, '>>%s\n', all_subs.names{i});                    
                    no_seqs = no_seqs_of_sub(obj, all_subs.names{i});
                    for j=1:length(no_seqs)
                        if (~ismember(no_seqs(j), no_used_seqs))                        
                            seq = obj.seqs(no_seqs(j));
                            fprintf(fid, '>%s\n', seq.name);
                            fprintf(fid, '%s\n', seq.genome(begin_sector:end_sector));
                        end
                    end
                end
            end
            fclose(fid);
        end
        
        % generate representative emission and transitios frequencies
        function gen_repr_freqs(obj)
            global consts;            
            begin_sector = 1;
            end_sector = obj.seq_len - 1;
            sub = 'F';
            min_num_res = 10;
            len = end_sector - begin_sector + 1;
            
            % calculate represantative emission freqs
            fid = FilesDirs.sec_fopen(FilesDirs.repr_em_freqs_filename, 'w');
            em_freqs = obj.em_freqs(begin_sector, end_sector, obj.no_seqs_of_sub(sub));
            str = '';
            num_used_pos = 0;
            for i=1:len
                if sum(em_freqs(i, :)) >= min_num_res
                    str = [str num2str(i + begin_sector - 2) ' '];
                    for j=1:consts.alphabet.num_nucl
                        str = [str num2str(em_freqs(i, j)) ' '];
                    end
                    str = [str char(10)];
                    num_used_pos = num_used_pos + 1;
                end
            end
            fprintf(fid, '%d\n', num_used_pos);
            fprintf(fid, '%s', str);
            fclose(fid);
            disp('calculated represantative emission freqs')
            
            % calculate represantative transition freqs
            fid = FilesDirs.sec_fopen(FilesDirs.repr_tr_freqs_filename, 'w');
            tr_freqs = obj.tr_freqs(begin_sector, end_sector, obj.no_seqs_of_sub(sub));
            num_used_values = 2;
            str = '';
            num_used_pos = 0;
            match_source_idx = 1;
            for i=1:len                
                if sum(tr_freqs(i, match_source_idx, 1:num_used_values)) >= min_num_res
                    str = [str num2str(i + begin_sector - 2) ' '];
                    for j=1:num_used_values
                        str = [str num2str(tr_freqs(i, match_source_idx, j)) ' '];
                    end
                    str = [str char(10)];
                    num_used_pos = num_used_pos + 1;
                end
            end
            fprintf(fid, '%d\n', num_used_pos);
            fprintf(fid, '%s', str);
            fclose(fid);
            disp('calculated represantative transition freqs')
        end        

        function print_calc_etp_al(obj, scenario)
            fid = fopen(FilesDirs.calc_etp_al_filename(scenario), 'w');
            if fid == -1 
                error(['Unable to open: ' FilesDirs.calc_etp_al_filename(scenario)]) 
            end      
            for i=1:obj.num_seqs
                fprintf(fid, '>%s\n', obj.seqs(i).sub);
                fprintf(fid, '%s\n', obj.seqs(i).genome);
            end
            fclose(fid);
        end                
    end
end
