classdef Math
    methods (Static)
        % logit
        function ret = logit(p)      
            ret = log(p / (1 - p));
        end
        
        % inverse logit
        function ret = inv_logit(x)      
            ret = exp(x) / (1 + exp(x));
        end        
        
        % random additive change
        function change = random_change(lower, upper)  
            is_ok = false;
            while ~is_ok
                change = (2 * rand(1) - 1) * 10^(randi(upper - lower + 1) + lower - 1); 
                if abs(change) > 10^lower
                    is_ok = true;
                end
            end
        end
        
        % random multiplicative change
        function factor = random_factor(upper)      
            d = rand(1) * upper;
            if (rand(1) > 0.5)
                factor = 1 + d;
            else
                factor = 1 - d / (1 + d);
            end
        end
        
        % perform mcmc to improve model paras
        function res = mcmc(init_model_para, num_steps, test_seqs, al)
            res = struct('model_paras', 0, 'scores', 0, 'sens', 0, 'spec', 0, 'is_valid', 0, 'is_accepted', 0, 'orig_segms', 0, 'est_segms', 0);
            BETA = 25;
            fid = FilesDirs.sec_fopen(FilesDirs.curr_mcmc_filename, 'w');
            fclose(fid);
            stamp = Stamp();
            
            % do initial scoring
            res_scoring = init_model_para.score(test_seqs, al);
            init_score = res_scoring.score;
            res.scores = init_score;
            res.sens = res_scoring.sens;
            res.spec = res_scoring.spec;
            res.model_paras = init_model_para;
            res.orig_segms = cell(num_steps + 1, test_seqs.num_seqs);
            res.est_segms = cell(num_steps + 1, test_seqs.num_seqs);
            res.seq_scores = zeros(num_steps + 1, test_seqs.num_seqs);
            res.seq_pos_scores = zeros(num_steps + 1, test_seqs.num_seqs);
            res.seq_segm_scores = zeros(num_steps + 1, test_seqs.num_seqs);
            for i=1:length(res_scoring.orig_segms)
                res.orig_segms{1, i} = res_scoring.orig_segms(i);
                res.est_segms{1, i} = res_scoring.est_segms(i);
                res.seq_scores(1, i) = res_scoring.seq_scores(i);
                res.seq_pos_scores(1, i) = res_scoring.seq_pos_scores(i);
                res.seq_segm_scores(1, i) = res_scoring.seq_segm_scores(i);
            end
            res.is_valid = true;
            res.is_accepted = true;
            
            % output to files
            FilesDirs.write2mcmc_files(stamp, 'Init');
            FilesDirs.write2mcmc_files(stamp, char(init_model_para));
            FilesDirs.write2mcmc_files(stamp, ['score: ' num2str(init_score)]);
            FilesDirs.write2mcmc_files(stamp, '');            
            
            no_step = 1;
            while (no_step <= num_steps)
                %try
                    disp(['performing MCMC step ' num2str(no_step)])
                    new_model_para = res.model_paras(no_step).change();
                    res_scoring = new_model_para.score(test_seqs, al);
                    new_score = res_scoring.score;
                    
                    if ((new_score / res.scores(no_step)) ^ BETA > rand(1))
                        res.model_paras(no_step + 1) = new_model_para;
                        res.scores(no_step + 1) = new_score;
                        res.sens = [res.sens res_scoring.sens];
                        res.spec = [res.spec res_scoring.spec];
                        for i=1:length(res_scoring.orig_segms)
                            res.orig_segms{no_step + 1, i} = res_scoring.orig_segms(i);
                            res.est_segms{no_step + 1, i} = res_scoring.est_segms(i);
                            res.seq_scores(no_step + 1, i) = res_scoring.seq_scores(i);
                            res.seq_pos_scores(no_step + 1, i) = res_scoring.seq_pos_scores(i);
                            res.seq_segm_scores(no_step + 1, i) = res_scoring.seq_segm_scores(i);
                        end
                        res.is_accepted = [res.is_accepted true];
                    else
                        res.model_paras(no_step + 1) = res.model_paras(no_step);
                        res.scores(no_step + 1) = res.scores(no_step);          
                        res.sens = [res.sens res.sens(no_step)];
                        res.spec = [res.spec res.spec(no_step)];
                        for i=1:length(res_scoring.orig_segms)
                            res.orig_segms{no_step + 1, i} = res.orig_segms{no_step, i};
                            res.est_segms{no_step + 1, i} = res.est_segms{no_step, i};
                            res.seq_scores(no_step + 1, i) = res.seq_scores(no_step, i);
                            res.seq_pos_scores(no_step + 1, i) = res.seq_pos_scores(no_step, i);
                            res.seq_segm_scores(no_step + 1, i) = res.seq_segm_scores(no_step, i);
                        end
                        res.is_accepted = [res.is_accepted false];
                    end

                    % output to files
                    FilesDirs.write2mcmc_files(stamp, ['MCMC step ' num2str(no_step)]);
                    FilesDirs.write2mcmc_files(stamp, char(new_model_para));
                    FilesDirs.write2mcmc_files(stamp, ['score: ' num2str(new_score)]);
                    if res.is_accepted(no_step + 1)
                        str_acc = 'yes';
                    else
                        str_acc = 'no';
                    end
                    FilesDirs.write2mcmc_files(stamp, ['accepted: ' str_acc]);
                    FilesDirs.write2mcmc_files(stamp, char(13));
                    
                    res.is_valid = [res.is_valid true];                    
                %catch e
                %    disp(e)
                %    res.is_valid = [res.is_valid false];      
                %    res.is_accepted = [res.is_accepted false];
                %end
                no_step = no_step + 1;
            end            
            
            str = [char(13) char(13)];
            for i=1:num_steps+1   
                str = [str '---------------------------------  no step ' num2str(i - 1) '  ---------------------------------' char(13)];            
                for j=1:test_seqs.num_seqs
                    str = [str 'orig ' num2str(j) ': ' Segmentation.segm2str(res.orig_segms{i, j}, true) char(13)];
                    str = [str 'est  ' num2str(j) ': ' Segmentation.segm2str(res.est_segms{i, j}, true) char(13)];
                    str = [str 'score ' num2str(j) ': ' num2str(res.seq_scores(i, j)) ' (' num2str(res.seq_pos_scores(i, j)) ', ' num2str(res.seq_segm_scores(i, j)) ')' char(13)];
                    str = [str char(13)];             
                end
                str = [str char(13) char(13)];
            end
            FilesDirs.write2mcmc_files(stamp, str);
        end
    end
end
